/*******************************************************************************

    PERM() - Generate all the permutation of the characters in a string.

    PERM() is a straightforward C implementation of the permutation algorithm
    from Ellis Horowitz's and Sartaj Sahni's FUNDAMENTALS OF DATA STRUCTURES.
    The algorithm, coded in the book's SPARKS pseudo-programming language,
    is found in Chapter 1, in the subsection about recursion.

        procedure PERM (A, k, n)

            if k = n then [print (A); return]

            B <- A

            for i <- k to n do
                call INTERCHANGE (A, k, i)
                call PERM (A, k+1, n)
                A <- B
            end

        end PERM

    NOTE that the SPARKS implementation indexes the characters in a string
    over the range 1..N, rather than C's 0..N-1.  The initial call to PERM()
    for the example string, "abcd", in the book would be as follows:

        call PERM ('abcd', 1, 4)

    This C implementation uses the book's 1..N range, so the above call would
    be rendered in C as:

        PERM ("abcd", 1, 4) ;

    INTERCHANGE() is the only function that actually accesses the individual
    characters in the string, so it automatically subtracts one from the
    indices in order to access the correct characters in the range 0..N-1.

*******************************************************************************/


#include  <stdio.h>			/* Standard I/O definitions. */
#include  <stdlib.h>			/* Standard C Library definitions. */
#include  <string.h>			/* Standard C string functions. */


/*******************************************************************************
    INTERCHANGE() - swaps the I-th and J-th characters in a string.  NOTE that
        I and J are in the range 1..N, where N is the length of the string.
        INTERCHANGE() automatically adjusts the indices to swap the correct
        characters in the C string.
*******************************************************************************/

void  INTERCHANGE (

    char  *string,		/* String containing characters to swap. */
    size_t  i,			/* Index 1..N of character to swap. */
    size_t  j)			/* Index 1..N of character to swap. */

{
    char  temp = string[i-1] ;
    string[i-1] = string[j-1] ;
    string[j-1] = temp ;
    return ;
}


/*******************************************************************************
    PERM() - recursively generates and prints out all of the permutations of
        the characters in a string.  The initial call to PERM() for a string
        should be "PERM (string, 1, strlen (string)) ;".
*******************************************************************************/

void  PERM (

    const  char  *A,		/* String of characters to be permuted. */
    size_t  k,			/* Index 1..N at which to begin permutation. */
    size_t  n)			/* Index 1..N of last character in string. */

{

    if (k == n) {		/* Permutation branch complete? */
        printf ("%s\n", A) ;
        return ;
    }
		/* SPARKS code uses B to cache copy of original A (B <- A)
		   and then modifies A.  In C, leave A untouched and modify
		   B copy instead. */
    char  *B = strdup (A) ;

    for (size_t  i = k ;  i <= n ;  i++) {
        INTERCHANGE (B, k, i) ;
        PERM ((const char *) B, k+1, n) ;
        strcpy (B, A) ;		/* Using B for changes, so B <- A. */
    }

    free (B) ;

    return ;

}


/*******************************************************************************
    MAIN() - generates and prints all the permutations of the characters
        in a string.  The string can be specified as the first argument
        on the command line; if no argument is given, all the permutations
        of "abcd" are generated.
*******************************************************************************/

int  main (
    int  argc,
    char  *argv[])
{
    char  *A = (argc > 1) ? argv[1] : "abcd" ;
    size_t  n = strlen (A) ;

    if (n > 0)  PERM (A, 1, n) ;

    exit (0) ;

}
